"""
classifier-pipeline - this is a server side component that manipulates cptv
files and to create a classification model of animals present
Copyright (C) 2018, The Cacophony Project

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from os import path
import numpy as np
from PIL import Image, ImageDraw, ImageFont

from track.clip import Clip
from ml_tools import tools
from ml_tools.rectangle import Rectangle
import ml_tools.globals as globs
from ml_tools.mpeg_creator import MPEGCreator

# from track.region import Region
from ml_tools.imageprocessing import normalize

PREVIEW_RAW = "raw"

PREVIEW_CLASSIFIED = "classified"

PREVIEW_NONE = "none"

PREVIEW_TRACKING = "tracking"

PREVIEW_BOXES = "boxes"

PREVIEW_OPTIONS = [
    None,
    PREVIEW_NONE,
    PREVIEW_RAW,
    PREVIEW_CLASSIFIED,
    PREVIEW_TRACKING,
    PREVIEW_BOXES,
]


class Previewer:

    TRACK_COLOURS = [(255, 0, 0), (0, 255, 0), (255, 255, 0), (128, 255, 255)]
    FILTERED_COLOURS = [(128, 128, 128)]

    def __init__(self, config, preview_type):
        self.config = config
        self.colourmap = self._load_colourmap()

        # make sure all the required files are there
        self.track_descs = {}
        self.preview_type = preview_type
        self.debug = config.debug

    @classmethod
    def create_if_required(self, config, preview_type):
        if preview_type and not preview_type.lower() == PREVIEW_NONE:
            return Previewer(config, preview_type)

    def _load_colourmap(self):
        colourmap_path = self.config.previews_colour_map
        if not path.exists(colourmap_path):
            colourmap_path = tools.resource_path("colourmap.dat")
        return tools.load_colourmap(colourmap_path)

    def export_clip_preview(self, filename, clip: Clip, predictions=None):
        """
        Exports a clip showing the tracking and predictions for objects within the clip.
        """

        logging.info("creating clip preview %s", filename)
        # increased resolution of video file.
        # videos look much better scaled up
        if not clip.stats:
            logging.error("Do not have temperatures to use.")
            return

        if self.debug:
            footer = Previewer.stats_footer(clip.stats)
        if predictions and (
            self.preview_type == PREVIEW_CLASSIFIED
            or self.preview_type == PREVIEW_TRACKING
        ):
            self.create_track_descriptions(clip, predictions)

        if clip.stats.min_temp is None or clip.stats.max_temp is None:
            thermals = [frame.thermal for frame in clip.frame_buffer.frames]
            clip.stats.min_temp = np.amin(thermals)
            clip.stats.max_temp = np.amax(thermals)

        res_x = clip.res_x
        res_y = clip.res_y
        if self.preview_type == PREVIEW_TRACKING:
            res_x *= 2
            res_y *= 2

        mpeg = MPEGCreator(str(filename))
        frame_scale = 4
        for frame_number, frame in enumerate(clip.frame_buffer):
            if self.preview_type == PREVIEW_RAW:
                image = self.convert_and_resize(
                    frame.thermal, clip.stats.min_temp, clip.stats.max_temp, clip.type
                )
                draw = ImageDraw.Draw(image)
            elif self.preview_type == PREVIEW_TRACKING:
                image = self.create_four_tracking_image(
                    frame,
                    clip.stats.min_temp,
                    clip.stats.max_temp,
                    clip.type,
                    frame_scale,
                )
                draw = ImageDraw.Draw(image)
                self.add_tracks(
                    draw,
                    clip.tracks,
                    frame_number,
                    predictions,
                    scale=frame_scale,
                )

            elif self.preview_type == PREVIEW_BOXES:
                image = self.convert_and_resize(
                    frame.thermal,
                    clip.stats.min_temp,
                    clip.stats.max_temp,
                    clip.type,
                    frame_scale=frame_scale,
                )
                draw = ImageDraw.Draw(image)
                screen_bounds = Rectangle(0, 0, image.width, image.height)
                self.add_tracks(
                    draw,
                    clip.tracks,
                    frame_number,
                    colours=[(128, 255, 255)],
                    scale=frame_scale,
                )

            elif self.preview_type == PREVIEW_CLASSIFIED:
                image = self.convert_and_resize(
                    frame.thermal,
                    clip.stats.min_temp,
                    clip.stats.max_temp,
                    clip.type,
                    frame_scale=frame_scale,
                )
                draw = ImageDraw.Draw(image)
                screen_bounds = Rectangle(0, 0, image.width, image.height)
                self.add_tracks(
                    draw,
                    clip.tracks,
                    frame_number,
                    predictions,
                    screen_bounds,
                    scale=frame_scale,
                )
            if frame.ffc_affected:
                self.add_header(draw, image.width, image.height, "Calibrating ...")
            if self.debug and draw:
                self.add_footer(
                    draw,
                    image.width,
                    image.height,
                    footer,
                    frame.ffc_affected,
                    frame_number,
                )
                self.add_header(
                    draw, image.width, image.height, f"Frame {frame.frame_number}"
                )

            mpeg.next_frame(np.asarray(image))
            # we store the entire video in memory so we need to cap the frame count at some point.
            if frame_number > clip.frames_per_second * 60 * 10:
                break
        clip.frame_buffer.close_cache()
        mpeg.close()

    def create_individual_track_previews(self, filename, clip: Clip):
        # resolution of video file.
        # videos look much better scaled up
        filename_format = path.splitext(filename)[0] + "-{}.mp4"

        FRAME_SIZE = 4 * 48
        frame_width, frame_height = FRAME_SIZE, FRAME_SIZE

        for id, track in enumerate(clip.tracks):
            video_frames = []
            for region in track.bounds_history:
                frame = clip.frame_buffer.get_frame(region.frame_number)
                cropped = frame.crop_by_region(region)
                if cropped.thermal.size == 0:
                    continue
                img = tools.convert_heat_to_img(
                    cropped.thermal,
                    self.colourmap,
                    np.amin(cropped.thermal),
                    np.amax(cropped.thermal),
                )
                img = img.resize((frame_width, frame_height), Image.NEAREST)
                video_frames.append(np.asarray(img))

            logging.info("creating preview %s", filename_format.format(id + 1))
            tools.write_mpeg(filename_format.format(id + 1), video_frames)

    def convert_and_resize(
        self, frame, h_min, h_max, type, size=None, mode=Image.BILINEAR, frame_scale=1
    ):
        """Converts the image to colour using colour map and resize"""
        thermal = frame.copy()
        if type == "IR":
            thermal = thermal[..., np.newaxis]
            thermal = np.repeat(thermal, 3, axis=2)
            image = Image.fromarray(np.uint8(thermal))
        else:
            image = tools.convert_heat_to_img(frame, self.colourmap, h_min, h_max)

        # image.save("test" + ".thumbnail", "JPEG")
        if frame_scale != 1:
            image = image.resize(
                (
                    int(image.width * frame_scale),
                    int(image.height * frame_scale),
                ),
                mode,
            )

        if self.debug:
            tools.add_heat_number(image, thermal, frame_scale)
        return image

    def create_track_descriptions(self, clip, predictions):
        if predictions is None:
            return
        # look for any tracks that occur on this frame

        for track in clip.tracks:
            guesses = []
            for model_pred in predictions.values():
                guesses.extend(model_pred.guesses_for(track.get_id()))
            track_description = "\n".join(guesses)
            track_description.strip()
            self.track_descs[track] = track_description

    def add_regions(draw, regions, v_offset=0, scale=1):
        for rect in regions:
            draw.rectangle(
                rect_points(rect, v_offset, scale=scale), outline=(128, 128, 128)
            )

    def add_tracks(
        self,
        draw,
        tracks,
        frame_number,
        track_predictions=None,
        screen_bounds=None,
        colours=TRACK_COLOURS,
        tracks_text=None,
        v_offset=0,
        scale=1,
    ):
        # look for any tracks that occur on this frame
        for index, track in enumerate(tracks):
            frame_offset = frame_number - track.start_frame
            if frame_offset >= 0 and frame_offset < len(track.bounds_history):
                region = track.bounds_history[frame_offset]
                prediction = None
                if track_predictions:
                    # need to handle mulitple models better here
                    model_pred = list(track_predictions.values())[1]
                    prediction = model_pred.prediction_for(track.get_id())
                add_track(
                    draw,
                    track,
                    self.track_descs.get(track),
                    region,
                    frame_offset,
                    colours[index % len(colours)],
                    prediction,
                    screen_bounds,
                    tracks_text,
                    v_offset=v_offset,
                    debug=self.debug,
                    scale=scale,
                )

    def add_header(
        self,
        draw,
        width,
        height,
        text,
    ):
        font = get_font()
        footer_size = font.getsize(text)
        center = (width / 2 - footer_size[0] / 2.0, 5)
        draw.text((center[0], center[1]), text, font=font)

    def add_footer(self, draw, width, height, text, ffc_affected, frame_number):
        font = get_font()
        footer_text = "{} FFC {} {}".format(frame_number, ffc_affected, text)
        footer_size = font.getsize(footer_text)
        center = (width / 2 - footer_size[0] / 2.0, height - footer_size[1])
        draw.text((center[0], center[1]), footer_text, font=font)

    def create_four_tracking_image(self, frame, min_temp, max_temp, type, frame_scale):
        thermal = frame.thermal
        filtered = frame.filtered
        if type == "IR":
            filtered = np.uint8(filtered)
            thermal = thermal[..., np.newaxis]
            thermal = np.repeat(thermal, 3, axis=2)
            filtered[filtered > 0] = 255
            filtered = filtered[..., np.newaxis]
            filtered = np.repeat(filtered, 3, axis=2)
            thermal = Image.fromarray(thermal)
            filtered = Image.fromarray(filtered)
        else:
            filtered, _ = normalize(filtered, new_max=255)
            filtered = Image.fromarray(filtered)
            filtered = filtered.convert("RGB")
            thermal = tools.convert_heat_to_img(
                thermal, self.colourmap, min_temp, max_temp
            )
        if self.debug:
            tools.add_heat_number(thermal, frame.thermal, 1)
        if frame.mask is None:
            mask = np.zeros((np.array(thermal).shape), dtype=np.uint8)
            mask = Image.fromarray(mask)
        else:
            mask = frame.mask * 255
            # mask, _ = normalize(frame.mask, new_max=255)
            mask = np.uint8(mask)

            mask = mask[..., np.newaxis]
            mask = np.repeat(mask, 3, axis=2)
            mask = Image.fromarray(mask)

        flow_h, flow_v = frame.get_flow_split(clip_flow=True)
        if flow_h is None and flow_v is None:
            flow_magnitude = Image.fromarray(
                np.zeros((np.array(thermal).shape), dtype=np.uint8)
            )
        else:
            flow_magnitude = (
                np.linalg.norm(np.float32([flow_h, flow_v]), ord=2, axis=0) / 4.0
                + min_temp
            )
            flow_magnitude = tools.convert_heat_to_img(
                flow_magnitude, self.colourmap, min_temp, max_temp
            )

        image = np.hstack(
            (np.vstack((thermal, mask)), np.vstack((filtered, flow_magnitude)))
        )
        image = Image.fromarray(image)
        image = image.resize(
            (
                int(image.width * frame_scale),
                int(image.height * frame_scale),
            ),
            Image.BILINEAR,
        )
        return image

    @staticmethod
    def stats_footer(stats):
        return "max {}, min{}, mean{}, filtered deviation {}, avg delta{}, temp_thresh {}".format(
            none_or_round(stats.max_temp),
            none_or_round(stats.min_temp),
            none_or_round(stats.mean_temp),
            none_or_round(stats.filtered_deviation, 2),
            none_or_round(stats.average_delta, 1),
            none_or_round(stats.temp_thresh),
        )


def none_or_round(value, decimals=0):
    if value:
        return round(value, decimals)
    else:
        return value


def add_class_results(
    draw,
    track,
    track_desc,
    frame_offset,
    rect,
    track_prediction,
    screen_bounds,
    v_offset=0,
    scale=1,
):
    if track_prediction is None:
        return

    current_prediction_string = track_prediction.get_classified_footer(
        frame_offset + track.start_frame
    )
    add_text_to_track(
        draw,
        rect,
        track_desc,
        current_prediction_string,
        screen_bounds,
        v_offset,
        scale=scale,
    )


def add_text_to_track(
    draw, rect, header_text, footer_text, screen_bounds, v_offset=0, scale=1
):
    font = get_font()
    font_title = get_title_font()
    header_size = font_title.getsize(header_text)
    footer_size = font.getsize(footer_text)
    # figure out where to draw everything
    header_rect = Rectangle(
        rect.left * scale,
        (v_offset + rect.top) * scale - header_size[1],
        header_size[0],
        header_size[1],
    )
    footer_center = ((rect.width * scale) - footer_size[0]) / 2
    footer_rect = Rectangle(
        rect.left * scale + footer_center,
        (v_offset + rect.bottom) * scale,
        footer_size[0],
        footer_size[1],
    )

    fit_to_image(header_rect, screen_bounds)
    fit_to_image(footer_rect, screen_bounds)

    draw.text((header_rect.x, header_rect.y), header_text, font=font_title)
    draw.text((footer_rect.x, footer_rect.y), footer_text, font=font)


def fit_to_image(rect: Rectangle, screen_bounds: Rectangle):
    """Modifies rect so that rect is visible within bounds."""
    if screen_bounds is None:
        return
    if rect.left < screen_bounds.left:
        rect.x = screen_bounds.left
    if rect.top < screen_bounds.top:
        rect.y = screen_bounds.top

    if rect.right > screen_bounds.right:
        rect.x = screen_bounds.right - rect.width

    if rect.bottom > screen_bounds.bottom:
        rect.y = screen_bounds.bottom - rect.height


def add_track(
    draw,
    track,
    track_desc,
    region,
    frame_offset,
    colour,
    track_prediction=None,
    screen_bounds=None,
    tracks_text=None,
    v_offset=0,
    scale=1,
    debug=False,
):
    if region.width == 0 or region.height == 0:
        return
    draw.rectangle(rect_points(region, v_offset, scale=scale), outline=colour)
    if track_prediction:
        add_class_results(
            draw,
            track,
            track_desc,
            frame_offset,
            region,
            track_prediction,
            screen_bounds,
            v_offset=v_offset,
            scale=scale,
        )

    if debug:
        text = None
        if tracks_text and len(tracks_text) > index:
            text = tracks_text[index]
        add_debug_text(
            draw,
            track,
            region,
            screen_bounds,
            text=text,
            v_offset=v_offset,
            frame_offset=frame_offset,
            scale=scale,
        )


def get_font(font_size=20):
    """gets default font."""
    if not globs._previewer_font:
        globs._previewer_font = ImageFont.truetype(
            tools.resource_path("Ubuntu-R.ttf"), font_size
        )
    return globs._previewer_font


def get_title_font():
    """gets default title font."""
    if not globs._previewer_font_title:
        globs._previewer_font_title = ImageFont.truetype(
            tools.resource_path("Ubuntu-B.ttf"), 14
        )
    return globs._previewer_font_title


def rect_points(rect, v_offset=0, h_offset=0, scale=1):
    return [
        scale * (rect.left + h_offset),
        scale * (rect.top + v_offset),
        scale * (rect.right + h_offset) - 1,
        scale * (rect.bottom + v_offset) - 1,
    ]
    return


def add_last_frame_tracking(
    frame,
    tracks,
    track_predictions=None,
    screen_bounds=None,
    colours=Previewer.TRACK_COLOURS,
    tracks_text=None,
    v_offset=0,
    scale=1,
    debug=False,
):
    image = tools.convert_heat_to_img(frame.thermal)

    draw = ImageDraw.Draw(image)

    # look for any tracks that occur on this frame
    for index, track in enumerate(tracks):
        region = track.bounds_history[-1]
        if region.frame_number != frame.frame_number:
            continue
        draw.rectangle(
            rect_points(region, v_offset, scale=scale),
            outline=colours[index % len(colours)],
        )
        if track_predictions is not None:
            track_prediction = track_predictions.prediction_for(track)
            if track_prediction:
                footer_text = track_prediction.get_classified_footer(labels)
                add_text_to_track(
                    draw,
                    region,
                    str(track.get_id()),
                    footer_text,
                    screen_bounds,
                    v_offset,
                )

        if debug:
            text = None
            if tracks_text and len(tracks_text) > index:
                text = tracks_text[index]
            add_debug_text(
                draw, track, region, screen_bounds, text=text, v_offset=v_offset
            )
    return image


def add_debug_text(
    draw,
    track,
    region,
    screen_bounds,
    text=None,
    v_offset=0,
    frame_offset=0,
    scale=1,
):
    font = get_font()
    if text is None:
        text = "id {}".format(track.get_id())

        text += "mass {} var {} vel ({},{}) blank? {}".format(
            region.mass,
            round(region.pixel_variance, 2),
            track.vel_x[frame_offset],
            track.vel_y[frame_offset],
            region.blank,
        )
    footer_size = font.getsize(text)
    footer_center = ((region.width * scale) - footer_size[0]) / 2

    footer_rect = tools.Rectangle(
        region.right * scale - footer_center / 2.0,
        (v_offset + region.bottom) * scale,
        footer_size[0],
        footer_size[1],
    )
    fit_to_image(footer_rect, screen_bounds)

    draw.text((footer_rect.x, footer_rect.y), text, font=font)



from multiprocessing import Pool

import cv2
import gc
import json
import logging
import os.path
import time

import numpy as np

from classify.trackprediction import Predictions
from track.clip import Clip
from track.cliptrackextractor import ClipTrackExtractor
from track.irtrackextractor import IRTrackExtractor
from pathlib import Path
from ml_tools import tools
from ml_tools.previewer import Previewer
from track.track import Track

from classify.thumbnail import get_thumbnail_info, best_trackless_thumb


class TrackExtractor:
    """Generate tracks for CPTV files."""

    def __init__(self, config, cache_to_disk=None, retrack=False):
        """Create an instance of a clip classifier"""

        self.config = config
        self.worker_threads = max(1, config.worker_threads)
        self.retrack = retrack
        if cache_to_disk is None:
            self.cache_to_disk = self.config.classify.cache_to_disk
        else:
            self.cache_to_disk = cache_to_disk

    def get_meta_data(self, filename):
        """Reads meta-data for a given cptv file."""
        source_meta_filename = os.path.splitext(filename)[0] + ".txt"
        if os.path.exists(source_meta_filename):
            meta_data = tools.load_clip_metadata(source_meta_filename)

            tags = set()
            for record in meta_data["Tags"]:
                # skip automatic tags
                if record.get("automatic", False):
                    continue
                else:
                    tags.add(record["animal"])

            tags = list(tags)

            if len(tags) == 0:
                tag = "no tag"
            elif len(tags) == 1:
                tag = tags[0] if tags[0] else "none"
            else:
                tag = "multi"
            meta_data["primary_tag"] = tag
            return meta_data
        else:
            return None

    def extract(self, base, to_stdout=False):
        # IF passed a dir extract all cptv files, if a cptv just extract this cptv file
        base = Path(base)
        if not base.exists():
            logging.error("Could not find file or directory %s", base)
            return
        if base.is_file():
            extract_file(base, self.config, self.cache_to_disk, self.retrack, to_stdout)
            return
        data = []
        for folder_path, _, files in os.walk(base):
            for name in files:
                if os.path.splitext(name)[1] in [".mp4", ".avi", ".cptv"]:
                    full_path = os.path.join(folder_path, name)
                    data.append(full_path)
        with Pool(
            self.worker_threads,
            init_worker,
            (self.config, self.cache_to_disk, to_stdout, self.retrack),
        ) as pool:
            pool.map(extract_thread, data)


config = None
cache_to_disk = None
to_stdout = False
retrack = False


def get_output_file(input_filename):
    return os.path.splitext(input_filename)[0]


def init_worker(c, cache, to_sout, re):
    global config
    global cache_to_disk
    global retrack
    global to_stdout
    config = c
    cache_to_disk = cache
    to_stdout = to_sout
    retrack = re


def extract_thread(filename):
    """
    Process a file extracting tracks and identifying them.
    :param filename: filename to process
    :param enable_preview: if true an MPEG preview file is created.
    """
    global config
    global cache_to_disk
    global to_stdout
    global retrack
    extract_file(filename, config, cache_to_disk, retrack=retrack, to_stdout=to_stdout)


def extract_file(filename, config, cache_to_disk, retrack=False, to_stdout=False):

    filename = Path(filename)
    if not filename.is_file():
        raise Exception("File {} not found.".format(filename))
    logging.info("Tracking %s", filename)
    previewer = Previewer.create_if_required(config, config.classify.preview)
    extension = filename.suffix
    if extension == ".cptv":
        track_extractor = ClipTrackExtractor(
            config.tracking,
            config.use_opt_flow,
            cache_to_disk,
            verbose=config.verbose,
        )
        logging.info("Using cptv extractor")

    else:
        track_extractor = IRTrackExtractor(
            config.tracking,
            cache_to_disk,
            verbose=config.verbose,
            keep_frames=True,
            # False if previewer is None else True,
            tracking_alg="subsense",
        )

        logging.info("Using ir extractor")
    clip = Clip(track_extractor.config, filename)
    if extension == ".cptv":
        clip.frames_per_second = 9
    else:
        clip.frames_per_second = 10
    existing_metadata = None
    if filename.with_suffix(".txt").exists():
        existing_metadata = tools.load_clip_metadata(filename.with_suffix(".txt"))

    if retrack:
        logging.info("Retracking")
        clip.load_metadata(existing_metadata)

    success = track_extractor.parse_clip(clip)

    # clip, success, tracking_time = extract_tracks(filename, config, cache_to_disk)
    if not success:
        logging.error("Could not parse %s", filename)
        return

    if retrack:
        for track in clip.tracks:
            track.trim()
            track.set_end_s(clip.frames_per_second)
    meta_filename = filename.with_suffix(".txt")

    if previewer:
        mpeg_filename = filename.parent / f"{filename.stem}-tracking.mp4"
        previewer.export_clip_preview(mpeg_filename, clip)
    logging.info("saving meta data %s", meta_filename)

    save_metadata(
        existing_metadata, filename, meta_filename, clip, track_extractor, to_stdout
    )
    if cache_to_disk:
        clip.frame_buffer.remove_cache()

    return clip, track_extractor


def save_metadata(
    existing_metadata, filename, meta_filename, clip, track_extractor, to_stdout=False
):
    metadata = clip.get_metadata()
    for i, track in enumerate(clip.tracks):
        best_thumb, best_score = get_thumbnail_info(clip, track)
        if best_thumb is None:
            metadata["tracks"][i]["thumbnail"] = None
            continue
        thumbnail_info = {
            "region": best_thumb.region,
            "contours": best_thumb.contours,
            "median_diff": best_thumb.median_diff,
            "score": round(best_score),
        }
        metadata["tracks"][i]["thumbnail"] = thumbnail_info
    if len(clip.tracks) == 0:
        # if no tracks choose a clip thumb
        region = best_trackless_thumb(clip)
        metadata["thumbnail_region"] = region
    metadata["source"] = str(filename)
    metadata["tracking_time"] = round(track_extractor.tracking_time, 1)
    metadata["algorithm"] = {}
    metadata["algorithm"]["tracker_version"] = track_extractor.tracker_version
    metadata["algorithm"]["tracker_config"] = track_extractor.config.as_dict()

    if existing_metadata is not None:
        # merge new metadata with old, think tracks is all that should be removed first
        if "tracks" in existing_metadata:
            del existing_metadata["tracks"]
        if "Tracks" in existing_metadata:
            del existing_metadata["Tracks"]

        existing_metadata.update(metadata)
        metadata = existing_metadata
    if to_stdout:
        print(json.dumps(metadata, cls=tools.CustomJSONEncoder))
    else:
        with open(meta_filename, "w") as f:
            json.dump(metadata, f, indent=4, cls=tools.CustomJSONEncoder)



"""
Processes a CPTV file identifying and tracking regions of interest, and saving them in the 'trk' format.
"""

import argparse
import os
import sys
from collections import namedtuple
from pathlib import Path

import h5py
import cv2
import numpy as np


def parse_params():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "target",
        default=None,
        help="h5py file to display",
    )
    args = parser.parse_args()
    args.target = Path(args.target)
    return args


Track = namedtuple("Track", "id regions what ai_what")


RESCALE = 4
from dateutil.parser import parse


def latest_date(dir):
    import json

    latest_date = None
    dbFiles = dir.glob("**/*.cptv")
    for dbName in dbFiles:
        meta_f = dbName.with_suffix(".txt")
        with open(meta_f, "r") as f:
            # add in some metadata stats
            meta = json.load(f)
        if meta.get("recordingDateTime"):
            meta["recordingDateTime"] = parse(meta["recordingDateTime"])
            if latest_date is None or meta["recordingDateTime"] > latest_date:
                latest_date = meta["recordingDateTime"]
    print("Latest date is", latest_date)


def makecsv(dir):
    dbFiles = dir.glob("**/*.hdf5")
    import csv

    print("Making csv")

    rec_counts = {}
    track_counts = {}
    with open("file-desc.csv", "w") as csvout:
        writer = csv.writer(csvout)
        writer.writerow(["file", "track id", "human tags", "extra human tags"])
        for dbName in dbFiles:
            f = h5py.File(dbName, "r")
            tracks = f["tracks"]
            track_ids = tracks.keys()
            has_track = False
            for track_id in track_ids:
                has_track = True
                human_tags = set()
                extra_tags = set()
                attrs = tracks[track_id].attrs
                if "human_tag" not in attrs:
                    continue
                tag = attrs["human_tag"]
                human_tags.add(tag)
                if "human_tags" in attrs:
                    extra = attrs["human_tags"]
                    for e in extra:
                        extra_tags.add(e)
                if tag in track_counts:
                    track_counts[tag] += 1
                else:
                    track_counts[tag] = 1
                writer.writerow([dbName, track_id, human_tags, extra_tags])
            if not has_track:
                writer.writerow([dbName, None, None, None])
            human_tags = list(human_tags)
            human_tags.sort()
            for t in human_tags:
                if t in rec_counts:
                    rec_counts[t] += 1
                else:
                    rec_counts[t] = 1

    print("REC Counts", rec_counts)
    print("Track Counts", track_counts)


def main():
    args = parse_params()
    latest_date(args.target)
    # makecsv(args.target)
    return
    f = h5py.File(args.target, "r")
    clip_attrs = f.attrs
    print(
        "Camera model is",
        clip_attrs.get("model", ""),
        " res:",
        clip_attrs["res_x"],
        clip_attrs["res_y"],
        " id: ",
        clip_attrs["clip_id"],
        " station id: ",
        clip_attrs["station_id"],
        " tags: ",
        clip_attrs.get("tags"),
    )
    for k in clip_attrs.keys():
        print(f"{k}: {clip_attrs[k]}")
    ffc_frames = clip_attrs["ffc_frames"]
    frames = f["frames"]
    if "background" in frames:
        background = frames["background"]
    else:
        print("No background found so using first frame")
        background = frames["thermals"][0]
    crop_rectangle = clip_attrs["crop_rectangle"]

    background = background[
        crop_rectangle[1] : crop_rectangle[3], crop_rectangle[0] : crop_rectangle[2]
    ]
    norm_back = norm_image(background)

    height, width, _ = norm_back.shape
    norm_back = cv2.resize(norm_back, (width * RESCALE, height * RESCALE))
    norm_back = cv2.putText(
        norm_back,
        "Background",
        (50, 50),
        cv2.FONT_HERSHEY_SIMPLEX,
        1,
        (0, 255, 0),
        2,
        cv2.LINE_AA,
    )

    cv2.imshow("b", norm_back)
    cv2.waitKey()
    cv2.destroyAllWindows()

    num_frames = clip_attrs["num_frames"]

    tracks_node = f["tracks"]
    tracks = []
    track_ids = tracks_node.keys()
    for track_id in track_ids:
        track_node = tracks_node[track_id]
        positions = track_node["regions"]
        id = track_node.attrs["id"]
        attrs = track_node.attrs
        human_tag = ""
        ai_tag = ""
        if "human_tag" in attrs:
            tag = attrs["human_tag"]
            conf = attrs["human_tag_confidence"]
            human_tag = f"{tag}-{conf}"

        if "ai_tag" in attrs:
            tag = attrs["ai_tag"]
            conf = attrs["ai_tag_confidence"]
            ai_tag = f"{tag}-{conf}"
        track = Track(id, positions, human_tag, ai_tag)
        tracks.append(track)
        print("For track", track_id)
        for k in attrs.keys():
            print(f"{k}: {attrs[k]}")
    frame_i = 0
    for frame in frames["thermals"]:
        frame = frame[
            crop_rectangle[1] : crop_rectangle[3], crop_rectangle[0] : crop_rectangle[2]
        ]

        filtered = np.float32(frame) - background
        filtered[filtered < 0] = 0
        filtered = norm_image(filtered)
        frame = norm_image(frame)

        height, width, _ = frame.shape
        frame = cv2.resize(frame, (width * RESCALE, height * RESCALE))
        filtered = cv2.resize(filtered, (width * RESCALE, height * RESCALE))

        if frame_i in ffc_frames:
            frame = cv2.putText(
                frame,
                "FFC",
                (50, 50),
                cv2.FONT_HERSHEY_SIMPLEX,
                1,
                (0, 255, 0),
                2,
                cv2.LINE_AA,
            )
            filtered = cv2.putText(
                filtered,
                "FFC",
                (50, 50),
                cv2.FONT_HERSHEY_SIMPLEX,
                1,
                (0, 255, 0),
                2,
                cv2.LINE_AA,
            )
        wait = 100
        for t in tracks:
            for r in t.regions:
                if r[4] == frame_i:
                    frame = cv2.rectangle(
                        frame,
                        (r[0] * 4, r[1] * 4),
                        (r[2] * 4, r[3] * 4),
                        (0, 255, 0),
                        1,
                    )
                    frame = cv2.putText(
                        frame,
                        f"human {t.what} \nai {t.ai_what}",
                        (r[0] * 4, r[1] * 4 + 20),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        0.5,
                        (0, 255, 0),
                        2,
                        cv2.LINE_AA,
                    )
                    filtered = cv2.rectangle(
                        filtered,
                        (r[0] * 4, r[1] * 4),
                        (r[2] * 4, r[3] * 4),
                        (0, 255, 0),
                        1,
                    )
                    filtered = cv2.putText(
                        filtered,
                        f"human {t.what} \nai {t.ai_what}",
                        (r[0] * 4, r[1] * 4 + 20),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        0.5,
                        (0, 255, 0),
                        2,
                        cv2.LINE_AA,
                    )
                    wait = 0
        cv2.imshow("thermal", frame)
        cv2.imshow("filtered", filtered)
        cv2.moveWindow("thermal", 0, 0)
        cv2.moveWindow("filtered", width * 5, 0)
        cv2.waitKey(wait)
        frame_i += 1
    print("DONE")
    cv2.waitKey(0)


def norm_image(image):
    max = np.amax(image)
    min = np.amin(image)
    norm = np.float32(image)
    norm = 255 * (norm - min) / (max - min)

    norm = np.uint8(norm)
    norm = norm[:, :, np.newaxis]
    norm = np.repeat(norm, 3, axis=2)
    return norm


if __name__ == "__main__":
    main()



"""
classifier-pipeline - this is a server side component that manipulates cptv
files and to create a classification model of animals present
Copyright (C) 2018, The Cacophony Project

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import datetime
import logging
import numpy as np
import os
import pytz
import cv2

from ml_tools.imageprocessing import normalize, detect_objects
from ml_tools.rectangle import Rectangle
from track.framebuffer import FrameBuffer
from track.track import Track
from track.region import Region
from piclassifier.cptvmotiondetector import is_affected_by_ffc

RES_X = 160
RES_Y = 120


class Clip:
    PREVIEW = "preview"
    FRAMES_PER_SECOND = 9
    local_tz = pytz.timezone("Pacific/Auckland")
    CLIP_ID = 1
    # used when calculating background, mininimum percentage the difference object
    # and background object must overlap i.e. they are a valid object
    MIN_ORIGIN_OVERLAP = 0.80

    def __init__(
        self,
        trackconfig,
        sourcefile,
        background=None,
        calc_stats=True,
        model=None,
        type="thermal",
        fps=FRAMES_PER_SECOND,
    ):
        self._id = Clip.CLIP_ID
        Clip.CLIP_ID += 1
        Track._track_id = 1
        self.disable_background_subtraction = False
        self.current_frame = -1
        self.ffc_affected = False
        self.crop_rectangle = None
        self.region_history = []
        self.active_tracks = set()
        self.tracks = []
        self.filtered_tracks = []
        self.from_metadata = False
        self.video_start_time = None
        self.location = None
        self.frame_buffer = None
        self.device = None
        self._background = None
        self.background_calculated = False
        self.res_x = None
        self.res_y = None
        self.background_frames = 0
        self.config = trackconfig
        self.frames_per_second = fps
        self.station_id = None
        self.calc_stats = calc_stats
        self.source_file = sourcefile
        self.stats = ClipStats()
        self.camera_model = None
        self.threshold_config = None
        self.track_min_delta = None
        self.track_max_delta = None
        self.background_thresh = None
        self.ffc_frames = []
        self.tags = None
        self.type = type
        # sets defaults
        self.set_model(model)
        if background is not None:
            self._background = background
            self._background_calculated()

        self.rescaled = None

    def get_frame(self, frame_number):
        return self.frame_buffer.get_frame(frame_number)

    def frames_kept(self):
        return self.frame_buffer.max_frames

    def rescaled_background(self, dims):
        if self.rescaled is not None:
            if self.rescaled[0] == self.current_frame:
                logging.info("Loading from cache")
                # 1 / 0
                return self.rescaled[1]
        resized = cv2.resize(
            self.background,
            (dims),
        )
        self.rescaled = (self.current_frame, resized)
        return resized

    @property
    def background(self):
        return self._background

    def set_model(self, camera_model):
        logging.debug("set model %s", camera_model)
        self.camera_model = camera_model
        threshold = self.config.motion.threshold_for_model(camera_model)
        if threshold:
            self.threshold_config = threshold
            self.set_motion_thresholds(threshold)

    def set_motion_thresholds(self, threshold):
        logging.debug("set thresholds %s", threshold)
        self.background_thresh = threshold.background_thresh
        self.temp_thresh = threshold.temp_thresh
        self.stats.threshold = self.background_thresh
        self.track_min_delta = threshold.track_min_delta
        self.track_max_delta = threshold.track_max_delta

    def _background_calculated(self):
        if self.type != "IR" or self.calc_stats:
            self.stats.mean_background_value = np.average(self._background)
        self.background_calculated = True

    def on_preview(self):
        return not self.background_calculated

    def set_background(self, frame):
        self._background = frame
        self._background_calculated()

    def update_background(self, frame):
        """updates the clip _background"""
        if self._background is None:
            self._background = frame
        else:
            self._background = np.minimum(self._background, frame)
        self.background_frames += 1

    def calculate_initial_diff(self, frame, initial_frames, initial_diff):
        """Compare this frame with the initial frames (to detect movement) and update the initial_diff
        frame (such that it is the maximum difference). This is essentially creating a frame which has
        all the warms moving parts from the initial frame
        """
        if initial_frames is None:
            return np.zeros((frame.shape))
        else:
            diff = initial_frames - frame
            if initial_diff is not None:
                initial_diff = np.maximum(initial_diff, diff)
            else:
                initial_diff = diff
        return initial_diff

    def calculate_background(self, frame_reader):
        """
        Calculate background by reading whole clip and grouping into sets of
        9 frames. Take the average of these 9 frames and use the minimum
        over the sets as the initial background
        Also check for animals in the background by checking for connected components in
        the intital_diff frame - this is the maximum change between first average frame and all other average frames in the clip
        """
        frame = frame_reader.next_frame()
        self.update_background(frame.pix)
        self._background_calculated()
        return

        # i dont think this is nessesary
        # first_frame = frame
        # initial_frames = None
        # initial_diff = None
        # frames = [frame.pix]
        # while True:
        #     frame = frame_reader.next_frame()
        #     if frame is None:
        #         break
        #     ffc_affected = is_affected_by_ffc(frame)
        #     if ffc_affected:
        #         continue
        #     frames.append(frame.pix)
        #     if len(frames) == 9:
        #         frame_average = np.average(frames, axis=0)
        #         self.update_background(frame_average)
        #         initial_diff = self.calculate_initial_diff(
        #             frame_average, initial_frames, initial_diff
        #         )
        #         if initial_frames is None:
        #             initial_frames = frame_average

        #         frames = []
        # if len(frames) > 0:
        #     frame_average = np.average(frames, axis=0)
        #     if initial_frames is None:
        #         initial_frames = frame_average
        #     self.update_background(frame_average)
        #     initial_diff = self.calculate_initial_diff(
        #         frame_average, initial_frames, initial_diff
        #     )

        #     if initial_frames is None:
        #         initial_frames = frame_average
        # frames = []
        # if initial_diff is None:
        #     if first_frame is not None:
        #         # fall back if whole clip is ffc
        #         self.update_background(frame.pix)
        #         self._background_calculated()
        #     return
        # np.clip(initial_diff, 0, None, out=initial_diff)
        # initial_frames = self.remove_background_animals(initial_frames, initial_diff)

        # self.update_background(initial_frames)
        # self._background_calculated()

    def remove_background_animals(self, initial_frame, initial_diff):
        """
        Try and remove animals that are already in the initial frames, by
        checking for connected components in the intital_diff frame
        (this is the maximum change between first frame and all other frames in the clip)
        """
        # remove some noise
        initial_diff[initial_diff < self.background_thresh] = 0
        initial_diff[initial_diff > 255] = 255
        initial_diff = np.uint8(initial_diff)
        initial_diff = cv2.fastNlMeansDenoising(initial_diff, None)

        _, lower_mask, lower_objects, centroids = detect_objects(
            initial_diff, otsus=True
        )

        max_region = Rectangle(0, 0, self.res_x, self.res_y)
        for component, centroid in zip(lower_objects[1:], centroids[1:]):
            region = Region(
                component[0],
                component[1],
                component[2],
                component[3],
                centroid=centroid,
            )
            region.enlarge(2, max=max_region)
            if region.width >= self.res_x or region.height >= self.res_y:
                logging.info(
                    "Background animal bigger than max, probably false positive %s %s",
                    region,
                    component[4],
                )
                continue
            background_region = region.subimage(initial_frame)
            norm_back = background_region.copy()
            norm_back, _ = normalize(norm_back, new_max=255)
            sub_components, sub_connected, sub_stats, centroids = detect_objects(
                np.uint8(norm_back), otsus=True
            )

            if sub_components <= 1:
                continue
            overlap_image = region.subimage(lower_mask) * 255
            overlap_pixels = np.sum(sub_connected[overlap_image > 0])
            overlap_pixels = overlap_pixels / float(component[4])

            # filter out components which are too big, or dont match original causes
            # for filtering
            if (
                overlap_pixels < Clip.MIN_ORIGIN_OVERLAP
                or sub_stats[1][4] == 0
                or sub_stats[1][4] == region.area
            ):
                logging.info(
                    "Invalid components mass: %s, components: %s region area %s overlap %s",
                    sub_stats[1][4],
                    sub_components,
                    region.area,
                    overlap_pixels,
                )
                continue

            sub_connected[sub_connected > 0] = 1
            # remove this component from the background by painting with
            # colours of neighbouring pixels
            background_region[:] = cv2.inpaint(
                np.float32(background_region),
                np.uint8(sub_connected),
                3,
                cv2.INPAINT_TELEA,
            )
        return initial_frame

    def _add_active_track(self, track):
        self.active_tracks.add(track)
        self.tracks.append(track)

    def get_id(self):
        return str(self._id)

    # def set_temp_thresh(self):
    #     if self.config.motion.dynamic_thresh:
    #         min_temp = self.threshold_config.min_temp_thresh
    #         max_temp = self.threshold_config.max_temp_thresh
    #         if max_temp:
    #             self.temp_thresh = min(max_temp, self.stats.mean_background_value)
    #         else:
    #             self.temp_thresh = self.stats.mean_background_value
    #         if min_temp:
    #             self.temp_thresh = max(min_temp, self.temp_thresh)
    #         self.stats.temp_thresh = self.temp_thresh
    #     else:
    #         self.temp_thresh = self.config.motion.temp_thresh

    def set_video_stats(self, video_start_time):
        """
        Extracts useful statics from video clip.
        """
        self.video_start_time = video_start_time
        self.stats.date_time = video_start_time.astimezone(Clip.local_tz)
        self.stats.is_night = (
            video_start_time.astimezone(Clip.local_tz).time().hour >= 2
        )

    def load_metadata(self, metadata, tag_precedence=None):
        self._id = metadata.get("id", 0)
        device_meta = metadata.get("Device")
        self.tags = metadata.get("Tags")

        if device_meta:
            self.device = device_meta.get("devicename")
        else:
            self.device = os.path.splitext(os.path.basename(self.source_file))[0].split(
                "-"
            )[-1]
        self.location = metadata.get("location")
        self.station_id = metadata.get("stationId")
        tracks = self.load_tracks_meta(metadata, tag_precedence)
        self.from_metadata = True
        self.tracks = set(tracks)

    def load_tracks_meta(self, metadata, tag_precedence):
        if "Tracks" in metadata:
            tracks_meta = metadata.get("Tracks", [])
        else:
            tracks_meta = metadata.get("tracks", [])
        tracks = []
        # get track data
        for track_meta in tracks_meta:
            track = Track(self.get_id())
            if track.load_track_meta(
                track_meta,
                self.frames_per_second,
                tag_precedence,
                self.config.min_tag_confidence,
            ):
                tracks.append(track)
        return tracks

    def start_and_end_in_secs(self, track):
        if track.end_s is None:
            track.end_s = (track.end_frame + 1) / self.frames_per_second

        return (track.start_s, track.end_s)

    def set_frame_buffer(
        self, high_quality_flow, cache_to_disk, use_flow, keep_frames, max_frames=None
    ):
        self.frame_buffer = FrameBuffer(
            self.source_file,
            high_quality_flow,
            cache_to_disk,
            use_flow,
            keep_frames,
            max_frames,
        )

    def set_res(self, res_x, res_y):
        if res_x == 0 or res_x == None:
            self.res_x = RES_X
        else:
            self.res_x = res_x
        if res_y == 0 or res_y == None:
            self.res_y = RES_Y
        else:
            self.res_y = res_y
        self._set_crop_rectangle()
        for track in self.tracks:
            track.crop_rectangle = self.crop_rectangle

    def _set_crop_rectangle(self):
        edge = self.config.edge_pixels
        self.crop_rectangle = Rectangle(
            edge, edge, self.res_x - 2 * edge, self.res_y - 2 * edge
        )

    def start_and_end_time_absolute(self, start_s=0, end_s=None):
        if not end_s:
            end_s = len(self.frame_buffer.frames) / self.frames_per_second
        return (
            self.video_start_time + datetime.timedelta(seconds=start_s),
            self.video_start_time + datetime.timedelta(seconds=end_s),
        )

    def print_if_verbose(self, info_string):
        if self.config.verbose:
            logging.info(info_string)

    def add_frame(self, thermal, filtered, mask=None, ffc_affected=False):
        self.current_frame += 1
        if ffc_affected:
            self.ffc_frames.append(self.current_frame)

        f = self.frame_buffer.add_frame(
            thermal, filtered, mask, self.current_frame, ffc_affected
        )

        if self.calc_stats:
            self.stats.add_frame(thermal, filtered)
        return f

    def get_metadata(self, predictions_per_model=None):
        meta_data = {}
        if self.camera_model:
            meta_data["camera_model"] = self.camera_model
        meta_data["background_thresh"] = self.background_thresh
        start, end = self.start_and_end_time_absolute()
        meta_data["start_time"] = start.isoformat()
        meta_data["end_time"] = end.isoformat()

        tracks = []
        for track in self.tracks:
            track_info = track.get_metadata(predictions_per_model)
            tracks.append(track_info)
        meta_data["tracks"] = tracks

        return meta_data


class ClipStats:
    """Stores background analysis statistics."""

    def __init__(self):
        self.mean_background_value = 0
        self.max_temp = None
        self.min_temp = None
        self.mean_temp = None
        self.frame_stats_min = []
        self.frame_stats_max = []
        self.frame_stats_median = []
        self.frame_stats_mean = []
        self.filtered_deviation = None
        self.filtered_sum = 0
        self.temp_thresh = 0
        self.threshold = None
        self.average_delta = None
        self.is_static_background = None

    def add_frame(self, thermal, filtered=None):
        f_median = np.median(thermal)
        f_max = np.max(thermal)
        f_min = np.min(thermal)
        f_mean = np.nanmean(thermal)
        self.max_temp = null_safe_compare(self.max_temp, f_max, max)
        self.min_temp = null_safe_compare(self.min_temp, f_min, min)

        self.frame_stats_min.append(f_min)
        self.frame_stats_max.append(f_max)
        self.frame_stats_median.append(f_median)
        self.frame_stats_mean.append(f_mean)
        if filtered is not None:
            self.filtered_sum += np.sum(np.abs(filtered))

    def completed(self):
        if self.filtered_sum is not None:
            self.filtered_deviation = np.mean(np.uint16(self.filtered_sum))
        self.mean_temp = np.mean(np.uint16(self.frame_stats_mean))


def null_safe_compare(a, b, cmp_fn):
    if a is None:
        return b
    elif b is not None:
        return cmp_fn(a, b)
    else:
        return None



import time
import logging
from piclassifier.recorder import Recorder
from piclassifier.cptvrecorder import CPTVRecorder
from piclassifier.eventreporter import throttled_event
from datetime import datetime


class ThrottledRecorder(Recorder):
    def __init__(self, recorder, thermal_config, headers, on_recording_stopping):
        self.bucket_size = thermal_config.throttler.bucket_size * headers.fps
        self.throttling = False
        self.tokens = self.bucket_size
        self.recorder = recorder
        self.last_rec = None
        self.last_motion = None
        self.fps = headers.fps
        self.no_motion = thermal_config.throttler.no_motion
        self.max_throttling_seconds = (
            thermal_config.throttler.max_throttling_minutes * 60
        )
        self.min_recording = self.recorder.min_frames
        self.throttled_at = None
        self.constant_recorder = False

    def final_name(self):
        return self.recorder.final_name()

    @property
    def name(self):
        return self.recorder.name

    @property
    def filename(self):
        return self.recorder.filename

    @filename.setter
    def filename(self, a):
        self.recorder.filename = a

    @property
    def recording(self):
        return self.recorder.recording

    @recording.setter
    def recording(self, a):
        self.recorder.recording = a

    @property
    def write_until(self):
        return self.recorder.write_until

    @write_until.setter
    def write_until(self, a):
        self.recorder.write_until = a

    @property
    def min_frames(self):
        return self.recorder.min_frames

    def force_stop(self):
        if self.recorder.recording:
            self.last_rec = time.time()
        self.recorder.force_stop()

    def process_frame(self, movement_detected, cptv_frame, received_at):
        if movement_detected:
            self.last_motion = received_at

        was_recording = self.recorder.recording
        self.recorder.process_frame(movement_detected, cptv_frame, received_at)
        self.take_token(received_at)
        if was_recording and not self.recorder.recording:
            self.last_rec = received_at

        if self.throttling and self.recorder.recording:
            logging.info("Throttling recording")
            self.stop_recording(received_at)

    def update_tokens(self, frame_time):
        if self.last_motion is None:
            return

        update_from = self.last_motion
        if self.last_rec and self.last_rec > self.last_motion:
            update_from = self.last_rec

        since_motion = frame_time - update_from
        # if we have been throttled wait for no motion before adding any tokens back
        if self.throttling:
            since_throttle = frame_time - self.throttled_at
            logging.debug(
                "Updating tokens %s seconds since motion with no motion %s since throttle %s, max throttle %s",
                round(since_motion),
                self.no_motion,
                datetime.fromtimestamp(since_throttle),
                self.max_throttling_seconds,
            )
            since_motion -= self.no_motion

            if since_motion < 0:
                if (
                    self.max_throttling_seconds
                    and since_throttle >= self.max_throttling_seconds
                ):
                    self.tokens = self.recorder.min_frames // 2
                    logging.info(
                        "Giving a few free tokens %s has been %s seconds since motion",
                        self.tokens,
                        round(since_motion),
                    )
                else:
                    return
            else:
                self.tokens += since_motion * self.fps
                logging.debug(
                    "Updating tokens %s seconds since motion has earnt %s tokens",
                    round(since_motion),
                    since_motion * self.fps,
                )
        else:
            logging.debug(
                "Updating tokens %s seconds since motion has earnt %s tokens",
                round(since_motion),
                since_motion * self.fps,
            )
            self.tokens += since_motion * self.fps
        self.throttling = False
        self.throttled_at = None
        self.tokens = int(self.tokens)
        self.tokens = min(self.tokens, self.bucket_size)

    def new_recording(self, background_frame, preview_frames, temp_thresh, frame_time):
        logging.debug("Attempting rec have %s tokens", self.tokens)
        self.update_tokens(frame_time)
        self.last_motion = frame_time
        if self.throttling:
            throttled_event()
            return False
        if self.tokens < self.min_recording:
            self.throttle(frame_time)
            return False
        self.take_token(frame_time, len(preview_frames))
        return self.recorder.new_recording(
            background_frame, preview_frames, temp_thresh, frame_time
        )

    def can_record(self, frame_time):
        prev_throttled = self.throttling
        self.update_tokens(frame_time)
        logging.debug(
            "Attempting rec have %s tokens was throttled %s throttled  %s tokens %s",
            self.tokens,
            prev_throttled,
            self.throttling,
            self.tokens,
        )
        self.last_motion = frame_time
        if self.throttling:
            if not prev_throttled:
                throttled_event()
            return False
        if self.tokens < self.min_recording:
            self.throttle(frame_time)
            return False
        return True

    def stop_recording(self, frame_time):
        if self.recorder.recording:
            self.last_rec = frame_time
            self.recorder.stop_recording(frame_time)

    def new_temp_name(self, frame_time):
        return self.recorder.new_temp_name(frame_time)

    def throttle(self, frame_time):
        logging.info("Throttling")
        self.throttling = True
        self.throttled_at = frame_time
        throttled_event()

    def take_token(self, frame_time, num_tokens=1):
        self.tokens -= num_tokens
        if self.tokens <= 0:
            self.tokens = 0
            self.throttle(frame_time)



"""
classifier-pipeline - this is a server side component that manipulates cptv
files and to create a classification model of animals present
Copyright (C) 2018, The Cacophony Project

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import time
import yaml
from datetime import datetime

from .clip import Clip
from piclassifier.cptvmotiondetector import is_affected_by_ffc
from ml_tools.imageprocessing import detect_objects, normalize
from track.cliptracker import ClipTracker
import logging
from cptv_rs_python_bindings import CptvReader
from piclassifier.motiondetector import WeightedBackground


class ClipTrackExtractor(ClipTracker):
    PREVIEW = "preview"
    VERSION = 11
    TYPE = "thermal"

    @property
    def tracker_version(self):
        return self.version

    @property
    def type(self):
        return ClipTrackExtractor.TYPE

    def __init__(
        self,
        config,
        use_opt_flow,
        cache_to_disk=False,
        keep_frames=True,
        calc_stats=True,
        high_quality_optical_flow=False,
        verbose=False,
        do_tracking=True,
        update_background=True,
        calculate_filtered=False,
        calculate_thumbnail_info=False,
        from_pi=False,
    ):
        super().__init__(
            config,
            cache_to_disk,
            keep_frames=keep_frames,
            calc_stats=calc_stats,
            verbose=verbose,
            do_tracking=do_tracking,
            calculate_thumbnail_info=calculate_thumbnail_info,
        )

        if from_pi:
            self.version = f"PI-{ClipTrackExtractor.VERSION}"
        else:
            self.version = ClipTrackExtractor.VERSION

        self.use_opt_flow = use_opt_flow
        self.high_quality_optical_flow = high_quality_optical_flow
        self.background_alg = None
        self.update_background = update_background
        self.calculate_filtered = calculate_filtered
        self.weighting_percent = 1
        # self.cache_to_disk = cache_to_disk
        # self.max_tracks = config.max_tracks
        # # frame_padding < 3 causes problems when we get small areas...
        # self.frame_padding = max(3, self.config.frame_padding)
        # # the dilation effectively also pads the frame so take it into consideration.
        # # self.frame_padding = max(0, self.frame_padding - self.config.dilation_pixels)
        # self.keep_frames = keep_frames
        # self.calc_stats = calc_stats
        # self._tracking_time = None
        # if self.config.dilation_pixels > 0:
        #     size = self.config.dilation_pixels * 2 + 1
        #     self.dilate_kernel = np.ones((size, size), np.uint8)

    def init_clip(self, clip):

        clip.set_frame_buffer(
            self.high_quality_optical_flow,
            self.cache_to_disk,
            self.use_opt_flow,
            self.keep_frames,
        )
        clip.type = self.type
        reader = CptvReader(str(clip.source_file))
        header = reader.get_header()

        clip.set_res(header.x_resolution, header.y_resolution)
        if clip.from_metadata:
            for track in clip.tracks:
                track.crop_regions()
        camera_model = None
        if header.model:
            camera_model = header.model
        clip.set_model(camera_model)

        # if we have the triggered motion threshold should use that
        # maybe even override dynamic threshold with this value
        if header.motion_config:
            motion = yaml.safe_load(header.motion_config)
            temp_thresh = motion.get("triggeredthresh")
            if temp_thresh:
                clip.temp_thresh = temp_thresh
        video_start_time = datetime.fromtimestamp(header.timestamp / 1000000)
        video_start_time = video_start_time.astimezone(Clip.local_tz)

        clip.set_video_stats(video_start_time)
        if camera_model == "lepton3.5":
            weight_add = 1 / self.weighting_percent
        else:
            weight_add = 0.1 / self.weighting_percent

        frame = reader.next_frame()
        clip.update_background(frame.pix)
        clip._background_calculated()
        self.background_alg = WeightedBackground(
            clip.crop_rectangle.x,
            clip.crop_rectangle,
            clip.res_x,
            clip.res_y,
            weight_add,
            clip.temp_thresh,
        )
        self.background_alg.process_frame(frame.pix)

    def parse_clip(self, clip, process_background=False):
        """
        Loads a cptv file, and prepares for track extraction.
        """

        self._tracking_time = None
        start = time.time()
        self.init_clip(clip)
        self._track_clip(clip, process_background=process_background)
        if self.calc_stats:
            clip.stats.completed()
        self._tracking_time = time.time() - start
        return True

    def _track_clip(self, clip, process_background=False):

        if clip.background is None:
            logging.error("Clip has no background have you called init_clip first")
            raise Exception("Clip has no background have you called init_clip first")
        reader = CptvReader(str(clip.source_file))
        while True:
            frame = reader.next_frame()

            if frame is None:
                break

            if not process_background and frame.background_frame:
                continue
            self.process_frame(clip, frame)
            if self.update_background or self.background_alg.background is None:
                # use mean of last 45 frames to update background, this will help
                # when pixels become cooler for a very short time i.e. tracked object is cooler than background
                last_avg = np.mean(
                    [f.thermal for f in clip.frame_buffer.get_last_x(x=45)], axis=0
                )
                self.background_alg.process_frame(last_avg)

        if not clip.from_metadata and self.do_tracking:
            self.apply_track_filtering(clip)

    @property
    def tracking_time(self):
        return self._tracking_time

    def start_tracking(
        self, clip, frames, track_frames=True, background_alg=None, **args
    ):
        # no need to retrack all of preview
        do_tracking = self.do_tracking
        self.background_alg = background_alg
        self.do_tracking = self.do_tracking and track_frames
        for frame in frames:
            self.process_frame(clip, frame)
        self.do_tracking = do_tracking

    def process_frame(self, clip, frame, **args):
        """
        Tracks objects through frame
        :param thermal: A numpy array of shape (height, width) and type uint16
        If specified background subtraction algorithm will be used.
        """
        ffc_affected = is_affected_by_ffc(frame)
        thermal = frame.pix.copy()
        if ffc_affected:
            self.print_if_verbose("{} ffc_affected".format(clip.current_frame))
        clip.ffc_affected = ffc_affected
        mask = None
        filtered = None
        if self.do_tracking or self.calculate_filtered or self.calculate_thumbnail_info:
            filtered = np.float32(frame.pix) - self.background_alg.background
        if self.do_tracking or self.calculate_thumbnail_info:
            obj_filtered, threshold = self._get_filtered_frame(
                clip, thermal, denoise=self.config.denoise
            )
            _, mask, component_details, centroids = detect_objects(
                obj_filtered, otsus=False, threshold=threshold, kernel=(5, 5)
            )
        cur_frame = clip.add_frame(thermal, filtered, mask, ffc_affected)
        if not self.do_tracking:
            return

        if clip.from_metadata:
            for track in clip.tracks:
                if clip.current_frame in track.frame_list:
                    track.add_frame_for_existing_region(
                        cur_frame,
                        threshold,
                        (
                            clip.frame_buffer.prev_frame.filtered
                            if clip.frame_buffer.prev_frame is not None
                            else None
                        ),
                    )
        else:
            regions = []
            if ffc_affected:
                clip.active_tracks = set()
            else:
                regions = self._get_regions_of_interest(
                    clip, component_details[1:], centroids[1:]
                )
                self._apply_region_matchings(clip, regions)
            clip.region_history.append(regions)



"""
classifier-pipeline - this is a server side component that manipulates cptv
files and to create a classification model of animals present
Copyright (C) 2018, The Cacophony Project

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

from abc import ABC, abstractmethod


class DefaultConfig(ABC):
    @classmethod
    @abstractmethod
    def get_defaults(cls):
        """The function to get default config."""
        ...

    @abstractmethod
    def validate(self):
        """The function to get default config."""
        ...


def deep_copy_map_if_key_not_exist(from_map, to_map):
    for key in from_map:
        if isinstance(from_map[key], dict):
            if key not in to_map:
                to_map[key] = {}
            deep_copy_map_if_key_not_exist(from_map[key], to_map[key])
        elif key not in to_map:
            to_map[key] = from_map[key]



# We store some cached shared objects as globals as they can not be passed around processes, and therefore would
# break the worker threads system.  Instead we load them on demand and store them in each processors global space.
_classifier = None
_previewer_font = None
_previewer_font_title = None



from collections import Counter
import tensorflow as tf
from functools import partial
import numpy as np
import logging
import random

AUTOTUNE = tf.data.AUTOTUNE


def get_weighting(
    dataset, labels, min_weigth=0.25, max_weight=4, excluded_labels=[], dont_weight=[]
):
    num_labels = len(labels)
    dist = get_distribution(dataset, num_labels)
    zeros = dist[dist == 0]
    non_zero_labels = num_labels - len(zeros)

    total = np.sum(dist)
    weights = {}
    for i in range(num_labels):
        if labels[i] in dont_weight:
            weights[i] = 1
        if dist[i] == 0:
            weights[i] = 0
        else:
            weights[i] = (1 / dist[i]) * (total / non_zero_labels)
            # cap the weights
            weights[i] = min(weights[i], 4)
            weights[i] = max(weights[i], 0.25)
        logging.info("weights for %s is %s", labels[i], weights[i])
    return weights


def get_distribution(dataset, num_labels, batched=True, one_hot=True, extra_meta=False):
    if extra_meta:
        true_categories = [y[0] for x, y in dataset]
    else:
        true_categories = [y for x, y in dataset]

    dist = np.zeros((num_labels), dtype=np.float32)
    if len(true_categories) == 0:
        return dist
    if batched:
        true_categories = tf.concat(true_categories, axis=0)
    if len(true_categories) == 0:
        return dist
    classes = []
    if one_hot:
        for y in true_categories:
            non_zero = tf.where(y).numpy()
            classes.extend(non_zero.flatten())
    else:
        classes = true_categories.flatten()
    classes = np.array(classes)

    c = Counter(list(classes))
    for i in range(num_labels):
        dist[i] = c[i]
    return dist


def get_dataset(load_function, base_dir, labels, **args):
    model_labels = args.get("model_labels")

    excluded_labels = args.get("excluded_labels", [])
    to_remap = args.get("remapped_labels", {})
    remapped = {}
    keys = []
    values = []
    shuffle_size = 4096
    if args.get("num_frames", 25) == 1:
        shuffle_size *= 20
    if model_labels is not None:
        new_labels = model_labels

        logging.info("Mapping DS labels %s to model labels %s", labels, model_labels)
        # if we are loading a model with different labels we need to map the dataset labels
        # to the equivalent model labels
        for l_i, og_lbl in enumerate(labels):
            keys.append(l_i)
            try:
                lbl = og_lbl
                if lbl in to_remap:
                    lbl = to_remap[lbl]

                mdl_i = model_labels.index(lbl)
                if lbl not in remapped:
                    remapped[lbl] = []
                remapped[lbl].append(og_lbl)
                values.append(mdl_i)
            except:
                remapped[og_lbl] = -1
                values.append(-1)

    else:

        logging.info("Excluding %s", excluded_labels)

        # get new labels after excluding and removing remapped labels
        new_labels = labels.copy()
        for excluded in excluded_labels:
            if excluded in new_labels:
                new_labels.remove(excluded)
        for remapped_lbl in to_remap.keys():
            if remapped_lbl in new_labels:
                new_labels.remove(remapped_lbl)

        # initialize remapped dictionary, setting labels that have been removed to -1, these values will be filtered later
        for l in labels:
            keys.append(labels.index(l))
            if l not in new_labels:
                remapped[l] = [-1]
                values.append(-1)
                logging.info("Excluding %s", l)
            else:
                remapped[l] = [l]
                values.append(new_labels.index(l))

        # add the remapped labels to the correct place
        for k, v in to_remap.items():
            if k in excluded_labels:
                continue
            if k in labels and v in new_labels:
                remapped[v].append(k)
                values[labels.index(k)] = new_labels.index(v)
                del remapped[k]

    remap_lookup = tf.lookup.StaticHashTable(
        initializer=tf.lookup.KeyValueTensorInitializer(
            keys=tf.constant(keys),
            values=tf.constant(values),
        ),
        default_value=tf.constant(-1),
        name="remapped_y",
    )
    num_labels = len(new_labels)
    logging.info("New labels are %s from original %s", new_labels, labels)
    for k, v in zip(keys, values):
        logging.info(
            "Mapping %s to %s", labels[k], new_labels[v] if v >= 0 else "nothing"
        )

    # 1 / 0
    filenames = tf.io.gfile.glob(f"{base_dir}/*.tfrecord")
    if not args.get("deterministic"):
        random.shuffle(filenames)

    dataset = load_function(filenames, remap_lookup, new_labels, args)
    if not args.get("one_hot", True):
        filter_excluded = lambda x, y: not tf.math.less(y, 0)
    else:
        if not args.get("include_track", False):
            filter_excluded = lambda x, y: not tf.math.equal(
                tf.math.count_nonzero(y), 0
            )
        else:
            filter_excluded = lambda x, y: not tf.math.equal(
                tf.math.count_nonzero(y[0]), 0
            )

    dataset = dataset.filter(filter_excluded)
    if dataset is None:
        logging.warn("No dataset for %s", filenames)
        return None, None

    if args.get("resample"):
        logging.info("RESAMPLING")
        # seems the only way to get even distribution
        label_ds = []
        unbalanced_ds = []
        dont_balance = ["vehicle"]
        for i, l in enumerate(new_labels):
            l_mask = np.zeros((len(new_labels)))
            l_mask[i] = 1
            # mask = tf.constant(mask, dtype=tf.float32)

            l_filter = lambda x, y: tf.math.reduce_all(tf.math.equal(y, l_mask))
            l_dataset = dataset.filter(l_filter)
            l_dataset = l_dataset.shuffle(shuffle_size, reshuffle_each_iteration=True)
            if l in dont_balance:
                unbalanced_ds.append(l_dataset)
            else:
                label_ds.append(l_dataset)
        dataset = tf.data.Dataset.sample_from_datasets(
            label_ds,
            # weights=[1 / len(new_labels)] * len(new_labels),
            stop_on_empty_dataset=True,
            rerandomize_each_iteration=True,
        )
        dont_balance.append(dataset)
        dataset = tf.data.Dataset.sample_from_datasets(
            dont_balance,
            # weights=[1 / len(new_labels)] * len(new_labels),
            stop_on_empty_dataset=False,
            rerandomize_each_iteration=True,
        )
    if args.get("epoch_size") is not None:
        dataset = dataset.take(args.get("epoch_size"))
        logging.info("Setting dataset to %s", args.get("epoch_size"))
    if args.get("cache", False):
        dataset = dataset.cache()
    if (
        not args.get("only_features")
        and args.get("shuffle", True)
        and not args.get("resample")
    ):
        logging.info("shuffling data with buffer %s", shuffle_size)
        dataset = dataset.shuffle(
            shuffle_size, reshuffle_each_iteration=args.get("reshuffle", True)
        )
    # tf refues to run if epoch sizes change so we must decide a costant epoch size even though with reject res
    # it will chang eeach epoch, to ensure this take this repeat data and always take epoch_size elements
    if not args.get("only_features"):
        dist = get_distribution(
            dataset,
            num_labels,
            batched=False,
            one_hot=args.get("one_hot", True),
            extra_meta=args.get("include_track", False),
        )
        for label, d in zip(new_labels, dist):
            logging.info("Have %s: %s", label, d)
        epoch_size = np.sum(dist)
        logging.info("Setting dataset size to %s", epoch_size)
        if not args.get("only_features", False):
            dataset = dataset.repeat(2)
        dataset = dataset.take(epoch_size)
        scale_epoch = args.get("scale_epoch", None)
        if scale_epoch:
            epoch_size = epoch_size // scale_epoch
            dataset = dataset.take(epoch_size)
    else:
        epoch_size = 1
    batch_size = args.get("batch_size", None)
    if batch_size is not None:
        dataset = dataset.batch(batch_size)
    dataset = dataset.prefetch(buffer_size=AUTOTUNE)

    return dataset, remapped, new_labels, epoch_size


def resample(dataset, labels):
    excluded_labels = ["sheep"]
    num_labels = len(labels)
    true_categories = [y for x, y in dataset]
    if len(true_categories) == 0:
        logging.info("no data")
        return None
    true_categories = np.int64(tf.argmax(true_categories, axis=1))
    c = Counter(list(true_categories))
    dist = np.empty((num_labels), dtype=np.float32)
    target_dist = np.empty((num_labels), dtype=np.float32)
    for i in range(num_labels):
        if labels[i] in excluded_labels:
            logging.info("Excluding %s for %s", c[i], labels[i])
            dist[i] = 0
        else:
            dist[i] = c[i]
            logging.info("Have %s for %s", dist[i], labels[i])
    zeros = dist[dist == 0]
    non_zero_labels = num_labels - len(zeros)
    target_dist[:] = 1 / non_zero_labels

    dist = dist / np.sum(dist)
    dist_max = np.max(dist)
    # really this is what we want but when the values become too small they never get sampled
    # so need to try reduce the large gaps in distribution
    # can use class weights to adjust more, or just throw out some samples
    max_range = target_dist[0] / 2
    for i in range(num_labels):
        if dist[i] == 0:
            target_dist[i] = 0
        elif dist_max - dist[i] > (max_range * 2):
            target_dist[i] = dist[i]

        target_dist[i] = max(0, target_dist[i])
    target_dist = target_dist / np.sum(target_dist)
